const inflection = require('inflection');
const stringUtils = require('ember-cli-string-utils');
const EOL = require('os').EOL;
const path = require('path');
const fs = require('fs');

module.exports = {
  description: 'Generates an ember-data model and the associated Mirage couterpart and screens.',

  anonymousOptions: ['name', 'attr:type'],

  fileMapTokens(options) {
    if (options.pod) {
      return {
        // Never tested. Not used to use pod structure ! 
        __root__() { return 'app'; },
        __path__() { return options.blueprintName;},
        __name__() { return options.dasherizedModuleName;},
        __mirage__() { return '../mirage'},
        __cloud__() { return '../cloud'}
      }
    } else {
      return {
        __root__() {
          if (options.inDummy) {
            return 'tests/dummy/app';
          } else {
            return 'app';
          } 
        },
        __path__() { return 'models';},
        __name__() { return options.dasherizedModuleName;},
        __mirage__() { return '../mirage'},
        __cloud__() { return '../cloud'}
      }
    }
  },

  locals(options) {
    let attrs = [], cols = [], needs = [], fact = [], form = [], imports = {}, ctrlVars= [], paramObject = [];
    let entityOptions = options.entity.options;
    let entityName = options.entity.name;
    let capitalizeEntityName = entityName.capitalize() ;

    const rel = Object.keys(entityOptions)
    .filter(key => (/has-many/.test(entityOptions[key]) || /belongs-to/.test(entityOptions[key])))
    .map(key => '\n\t\t' + stringUtils.camelize(key) + relFunction(entityOptions[key]))

    Object.values(entityOptions).forEach(value => {
      if (/has-many/.test(value)) {
        imports.hasMany = true
      }
      if (/belongs-to/.test(value)) {
        imports.belongsTo = true
      }
    })
 
    for (let name in entityOptions) {
      let type = entityOptions[name] || '';
      let foreignModel = name;
      if (type.indexOf(':') > -1) {
        foreignModel = type.split(':')[1];
        type = type.split(':')[0];
      }
      let dasherizedName = stringUtils.dasherize(name);
      let camelizedName = stringUtils.camelize(name);
      let dasherizedType = stringUtils.dasherize(type);
      let dasherizedForeignModel = stringUtils.dasherize(foreignModel);
      let dasherizedForeignModelSingular = inflection.singularize(dasherizedForeignModel);

      let attr;
      if (/has-many/.test(dasherizedType)) {
        let camelizedNamePlural = inflection.pluralize(camelizedName);
        attr = dsAttr(dasherizedForeignModelSingular, dasherizedType);
        attrs.push(camelizedNamePlural + ': ' + attr);
      } else if (/belongs-to/.test(dasherizedType)) {
        attr = dsAttr(dasherizedForeignModel, dasherizedType);
        attrs.push(camelizedName + ': ' + attr);
      } else {
        attr = dsAttr(dasherizedName, dasherizedType);
        attrs.push(camelizedName + ': ' + attr);
        cols.push("Ember.Object.create({\n\tpropertyName: \""+camelizedName+"\",\n\ttitle: \""+name+"\"\n}),");
      }

      if (/has-many|belongs-to/.test(dasherizedType)) {
        needs.push("'model:" + dasherizedForeignModelSingular + "'");
      }
      if (!/has-many/.test(dasherizedType) && !/belongs-to/.test(dasherizedType)) {
        fact.push('\n\t' + camelizedName + fakerize(dasherizedType));
        // todo Generate input form for creating entity (without any relationship)
        form.push('{{input type="text" name="'+camelizedName+'" value='+camelizedName+' placeholder="'+camelizedName+'"}}');
        ctrlVars.push(camelizedName+': null,');
      }
    }

    let needsDeduplicated = needs.filter(function(need, i) {
      return needs.indexOf(need) === i;
    });

    attrs = attrs.join(',' + EOL + '  ');
    needs = '  needs: [' + needsDeduplicated.join(', ') + ']';
    cols = cols.join(EOL);

    return {
      entity: entityName,
      attrs: attrs,
      cols: cols,
      needs: needs,
      fact: fact,
      rel: rel,
      ctrlVars: ctrlVars.join(EOL),
      form: form.join(EOL),
      imports: Object.keys(imports).map(relation => '\n\t' + relation)
    };
  },

  afterInstall(options) {
    let blueprintName = options.originBlueprintName;
    let entityName = options.entity.name;

    // Check that /mirage/serializers/applications.js is existing
    let mirageSerializerPath = (options.dummy) ? "tests/dummy/mirage/serializers/application.js" : "mirage/serializers/application.js";
    if (!fs.existsSync(mirageSerializerPath)) {
      this.ui.writeLine("Creating file "+mirageSerializerPath);
      let serializerContent = "import { JSONAPISerializer } from 'ember-cli-mirage';\n\n\texport default JSONAPISerializer.extend({\n});\n";
      fs.writeFileSync(mirageSerializerPath, serializerContent, 'utf-8', 'w+');
    }

    // Complete /mirage/scenarios/default.js
    let mirageScenarioPath = (options.dummy) ? "tests/dummy/mirage/scenarios/default.js" : "mirage/scenarios/default.js";
    let scenarioLine = '\tserver.createList\(\''+entityName+'\', 15\);\n';
    if (!fs.existsSync(mirageScenarioPath)) {
      this.ui.writeLine("Creating file "+mirageScenarioPath);
      let scenarioContent = "export default function(server) {\n"+scenarioLine+"}\n";
      fs.writeFileSync(mirageScenarioPath, scenarioContent, 'utf-8', 'w+');
    } else {
      addLineToFile(this, mirageScenarioPath, /export default function\(server\) {/, scenarioLine);
    }

    // Complete /mirage/config.js
    let mirageConfigPath = (options.dummy) ? "tests/dummy/mirage/config.js" : "mirage/config.js";
    let configLine = "this.get('/"+inflection.pluralize(entityName)+"', '"+inflection.pluralize(entityName)+"');\n"
    if (!fs.existsSync(mirageConfigPath)) {
      this.ui.writeLine("Creating file "+mirageConfigPath);
      let configContent = "export default function() {\n\tthis.namespace = '';\n"+configLine+"}\n";
      fs.writeFileSync(mirageConfigPath, configContent, 'utf-8','w+');
    } else {
      addLineToFile(this, mirageConfigPath, /this\.namespace = \'\';/, configLine);
    }

    // Add route in router.js for the entity page
    let routerPath = (options.dummy) ? "tests/dummy/app/router.js" : "app/router.js";
    let routeName = '\tthis.route\(\''+blueprintName+'/'+entityName+'\'\);';
    addLineToFile(this, routerPath, /Router\.map\(function\(\) {/, routeName);
  }
};

function addLineToFile(ctx, filePath, markerString, addedLine) {
  let fileContents = fs.readFileSync(filePath, 'utf-8');

  if (fileContents.indexOf(addedLine) === -1) {
   ctx.ui.writeLine("\tAdd line "+addedLine+" to file "+filePath);
   fileContents = fileContents.replace(markerString, [
     '$&',
     addedLine,
   ].join('\n'));
 }
 if (fileContents.indexOf(addedLine) === -1) {
   ctx.ui.writeWarnLine(
     'Unable to update '+filePath+'. You should update this file manually.');
 } else {
   fs.writeFileSync(filePath, fileContents, 'utf-8');
 }
}

function dsAttr(name, type) {
  switch (type) {
    case 'belongs-to':
      return "DS.belongsTo('" + name + "')";
    case 'has-many':
      return "DS.hasMany('" + name + "')";
    case '':
      //"If you don't specify the type of the attribute, it will be whatever was provided by the server"
      //https://emberjs.com/guides/models/defining-models/
      return 'DS.attr()';
    default:
      return "DS.attr('" + type + "')";
  }
}

const fakerize = (type) => {
  let typeFake
  switch (type) {
    case 'boolean':
      typeFake = 'faker.random.boolean()'
      break
    case 'date':
      typeFake = 'faker.date.past()'
      break
    case 'number':
      typeFake = 'faker.random.number({min: 0,max: 10000})'
      break
    case 'string':
      typeFake = 'faker.name.findName()'
      break
    default:
      return null
  }
  return ' () {\n\t\treturn ' + typeFake + '\n\t}'
}

const relFunction = (str) => {
  if (/has-many/.test(str)) {
    return ': hasMany()'
  }
  if (/belongs-to/.test(str)) {
    return ': belongsTo()'
  }
}